from __future__ import absolute_import

from django.apps import AppConfig
from django.core.signals import request_finished, request_started

from scout_apm.django.instruments.sql import SQLInstrument
from scout_apm.django.instruments.template import TemplateInstrument
from scout_apm.django.instruments.view import ViewInstrument
from scout_apm.django.config import ConfigAdapter
from scout_apm.core.tracked_request import TrackedRequest
import scout_apm.core

import logging


logger = logging.getLogger(__name__)


class DjangoSignals:
    @staticmethod
    def install():
        request_started.connect(DjangoSignals.start_tracked_request,
                                dispatch_uid='request_started_scoutapm')
        request_finished.connect(DjangoSignals.stop_tracked_request,
                                 dispatch_uid='request_stopped_scoutapm')
        logger.debug('Added Django Signals')

    # sender: django.core.handlers.wsgi.WSGIHandler
    # kwargs: 'environ' => { ENV Key => Env Value }
    #         'signal' => <django.dispatch.dispatcher.Signal object at 0x10ed7c470>
    def start_tracked_request(sender, **kwargs):
        # TODO: This is a good spot to extract headers
        operation = 'Django'
        tr = TrackedRequest.instance()
        tr.start_span(operation=operation)
        tr.mark_real_request()

    def stop_tracked_request(sender, **kwargs):
        TrackedRequest.instance().stop_span()


class ScoutApmDjangoConfig(AppConfig):
    name = 'scout_apm'
    verbose_name = 'Scout Apm (Django)'

    def ready(self):
        # Copy django configuration to scout_apm's config
        ConfigAdapter.install()

        # Finish installing the agent. If the agent isn't installed for any
        # reason, return without installing instruments
        installed = scout_apm.core.install()
        if installed is False:
            return


        # Setup Instruments
        DjangoSignals.install()
        SQLInstrument.install()
        TemplateInstrument.install()
        ViewInstrument.install()



# Python Built-Ins
from __future__ import absolute_import
import logging

# APM Modules
from scout_apm.core.monkey import CallableProxy, monkeypatch_method
from scout_apm.core.stacktracer import trace_function
from scout_apm.core.tracked_request import TrackedRequest

# Django
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIHandler
import django

try:
    from django.urls import resolvers
except:
    from django.core import urlresolvers


# Logging
logger = logging.getLogger(__name__)


def patch_function_list(functions, action_type, format_string):
    for i, func in enumerate(functions):
        if hasattr(func, 'im_class'):
            middleware_name = func.im_class.__name__
        else:
            middleware_name = func.__name__
        info = (action_type, {'name': middleware_name})
        functions[i] = trace_function(func, info)


def wrap_middleware_with_tracers(request_handler):
    # XXX: Figure out why request middleware isn't getting instrumented
    patch_function_list(request_handler._request_middleware, 'Middleware/Request', 'Middleware: %s (request)')
    patch_function_list(request_handler._view_middleware, 'Middleware/View', 'Middleware: %s (view)')
    patch_function_list(request_handler._template_response_middleware, 'Middleware/Template/Response', 'Middleware: %s (template response)')
    patch_function_list(request_handler._response_middleware, 'Middleware/Response', 'Middleware: %s (response)')
    patch_function_list(request_handler._exception_middleware, 'Middleware/Exception', 'Middleware: %s (exeption)')


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
middleware_patched = False


def intercept_middleware():
    @monkeypatch_method(WSGIHandler)
    def __call__(original, self, *args, **kwargs):
        # The middleware cache may have been built before we have a chance to monkey patch
        # it, so do so here
        global middleware_patched
        if not middleware_patched and self._request_middleware is not None:
            self.initLock.acquire()
            try:
                if not middleware_patched:
                    wrap_middleware_with_tracers(self)
                    middleware_patched = True
            finally:
                self.initLock.release()
        return original(*args, **kwargs)

    @monkeypatch_method(BaseHandler)
    def load_middleware(original, self, *args, **kwargs):
        global middleware_patched
        original(*args, **kwargs)
        wrap_middleware_with_tracers(self)
        middleware_patched = True


def intercept_resolver_and_view():
    # The only way we can really wrap the view method is by replacing the
    # implementation of RegexURLResolver.resolve. It would be nice if django
    # had more configurability here, but it does not.  However, we only want to
    # replace it when invoked directly from the request handling stack, so we
    # inspect the callstack in __new__ and return either a normal object, or an
    # instance of our proxying class.

    if django.VERSION > (2, 0):
        intercept_resolver_and_view_django_2()
    elif django.VERSION > (1, 9):
        intercept_resolver_and_view_django_1_9()
    else:
        intercept_resolver_and_view_django_1_8()


def intercept_resolver_and_view_django_2():
    real_resolver_cls = resolvers.URLResolver

    class ProxyURLResolverMetaClass(resolvers.URLResolver.__class__):
        def __instancecheck__(self, instance):
            # Some places in django do a type check against URLResolver
            # and behave differently based on the result, so we have to make
            # sure the replacement class we plug in accepts instances of both
            # the default and replaced types.
            return isinstance(instance, real_resolver_cls) or super(ProxyURLResolverMetaClass, self).__instancecheck__(instance)

    class ProxyURLResolver(object):
        __metaclass__ = ProxyURLResolverMetaClass

        def __new__(cls, *args, **kwargs):
            real_object = real_resolver_cls(*args, **kwargs)
            obj = super(ProxyURLResolver, cls).__new__(cls)
            obj.other = real_object
            return obj
            # XXX: this return is behind an if statement in speedbar, sometimes
            # doesn't instrument. unsure why

        def __getattr__(self, attr):
            return getattr(self.other, attr)

        def resolve(self, path):
            callbacks = self.other.resolve(path)
            callbacks.func = self.trace_view_function(callbacks.func, ('Controller', {"path": path, "name": callbacks._func_path}))
            return callbacks

        # XXX: This is duplicate code w/ StackTracer class.  Can this maybe be
        # a callback that takes the span, and the req and the args?  Rather
        # than a full reimpl of this?
        def trace_view_function(self, func, info):
            try:
                def tracing_function(original, *args, **kwargs):
                    entry_type, detail = info

                    operation = entry_type
                    if detail['name'] is not None:
                        operation = operation + '/' + detail['name']

                    span = TrackedRequest.instance().start_span(operation=operation)
                    for key in detail:
                        span.tag(key, detail[key])

                    # And the custom View stuff
                    request = args[0]

                    # Extract headers
                    #  regex = re.compile('^HTTP_')
                    #  headers = dict((regex.sub('', header), value) for (header, value)
                    #  in request.META.items() if header.startswith('HTTP_'))

                    span.tag('remote_addr', request.META['REMOTE_ADDR'])

                    logger.debug('Before calling original view')
                    try:
                        return original(*args, **kwargs)
                    except Exception as e:
                        logger.debug('***** Got the exception')
                        TrackedRequest.instance().tag('error', 'true')
                        raise e
                    finally:
                        TrackedRequest.instance().stop_span()
                        logger.debug(span.dump())

                return CallableProxy(func, tracing_function)
            except Exception as err:
                logger.debug(err)
                # If we can't wrap for any reason, just return the original
                return func
    resolvers.URLResolver = ProxyURLResolver


def intercept_resolver_and_view_django_1_9():
    real_resolver_cls = resolvers.RegexURLResolver

    class ProxyRegexURLResolverMetaClass(resolvers.RegexURLResolver.__class__):
        def __instancecheck__(self, instance):
            # Some places in django do a type check against RegexURLResolver
            # and behave differently based on the result, so we have to make
            # sure the replacement class we plug in accepts instances of both
            # the default and replaced types.
            return isinstance(instance, real_resolver_cls) or super(ProxyRegexURLResolverMetaClass, self).__instancecheck__(instance)

    class ProxyRegexURLResolver(object):
        __metaclass__ = ProxyRegexURLResolverMetaClass

        def __new__(cls, *args, **kwargs):
            real_object = real_resolver_cls(*args, **kwargs)
            obj = super(ProxyRegexURLResolver, cls).__new__(cls)
            obj.other = real_object
            return obj
            # XXX: this return is behind an if statement in speedbar, sometimes
            # doesn't instrument. unsure why

        def __getattr__(self, attr):
            return getattr(self.other, attr)

        def resolve(self, path):
            callbacks = self.other.resolve(path)
            callbacks.func = self.trace_view_function(callbacks.func, ('Controller', {"path": path, "name": callbacks._func_path}))
            return callbacks

        # XXX: This is duplicate code w/ StackTracer class.  Can this maybe be
        # a callback that takes the span, and the req and the args?  Rather
        # than a full reimpl of this?
        def trace_view_function(self, func, info):
            try:
                def tracing_function(original, *args, **kwargs):
                    entry_type, detail = info

                    operation = entry_type
                    if detail['name'] is not None:
                        operation = operation + '/' + detail['name']

                    span = TrackedRequest.instance().start_span(operation=operation)
                    for key in detail:
                        span.tag(key, detail[key])

                    # And the custom View stuff
                    request = args[0]

                    # Extract headers
                    #  regex = re.compile('^HTTP_')
                    #  headers = dict((regex.sub('', header), value) for (header, value)
                    #  in request.META.items() if header.startswith('HTTP_'))

                    span.tag('remote_addr', request.META['REMOTE_ADDR'])

                    logger.debug('Before calling original view')
                    try:
                        return original(*args, **kwargs)
                    except Exception as e:
                        logger.debug('***** Got the exception')
                        TrackedRequest.instance().tag('error', 'true')
                        raise e
                    finally:
                        TrackedRequest.instance().stop_span()
                        logger.debug(span.dump())

                return CallableProxy(func, tracing_function)
            except Exception as err:
                logger.debug(err)
                # If we can't wrap for any reason, just return the original
                return func

    resolvers.RegexURLResolver = ProxyRegexURLResolver


def intercept_resolver_and_view_django_1_8():
    real_resolver_cls = urlresolvers.RegexURLResolver

    class ProxyRegexURLResolverMetaClass(urlresolvers.RegexURLResolver.__class__):
        def __instancecheck__(self, instance):
            # Some places in django do a type check against RegexURLResolver
            # and behave differently based on the result, so we have to make
            # sure the replacement class we plug in accepts instances of both
            # the default and replaced types.
            return isinstance(instance, real_resolver_cls) or super(ProxyRegexURLResolverMetaClass, self).__instancecheck__(instance)

    class ProxyRegexURLResolver(object):
        __metaclass__ = ProxyRegexURLResolverMetaClass

        def __new__(cls, *args, **kwargs):
            real_object = real_resolver_cls(*args, **kwargs)
            obj = super(ProxyRegexURLResolver, cls).__new__(cls)
            obj.other = real_object
            return obj
            # XXX: this return is behind an if statement in speedbar, sometimes
            # doesn't instrument. unsure why

        def __getattr__(self, attr):
            return getattr(self.other, attr)

        def resolve(self, path):
            callbacks = self.other.resolve(path)
            callbacks.func = self.trace_view_function(callbacks.func, ('Controller', {"path": path, "name": callbacks._func_path}))
            return callbacks

        # XXX: This is duplicate code w/ StackTracer class.  Can this maybe be
        # a callback that takes the span, and the req and the args?  Rather
        # than a full reimpl of this?
        def trace_view_function(self, func, info):
            try:
                def tracing_function(original, *args, **kwargs):
                    entry_type, detail = info

                    operation = entry_type
                    if detail['name'] is not None:
                        operation = operation + '/' + detail['name']

                    span = TrackedRequest.instance().start_span(operation=operation)
                    for key in detail:
                        span.tag(key, detail[key])

                    # And the custom View stuff
                    request = args[0]

                    # Extract headers
                    #  regex = re.compile('^HTTP_')
                    #  headers = dict((regex.sub('', header), value) for (header, value)
                    #  in request.META.items() if header.startswith('HTTP_'))

                    span.tag('remote_addr', request.META['REMOTE_ADDR'])

                    logger.debug('Before calling original view')
                    try:
                        return original(*args, **kwargs)
                    except Exception as e:
                        logger.debug('***** Got the exception')
                        TrackedRequest.instance().tag('error', 'true')
                        raise e
                    finally:
                        TrackedRequest.instance().stop_span()
                        logger.debug(span.dump())

                return CallableProxy(func, tracing_function)
            except Exception as err:
                logger.debug(err)
                # If we can't wrap for any reason, just return the original
                return func

    urlresolvers.RegexURLResolver = ProxyRegexURLResolver

class ViewInstrument:
    @staticmethod
    def install():
        intercept_middleware()
        intercept_resolver_and_view()
        logger.debug('Monkey patched View')



import logging
import psutil

# Logging
logger = logging.getLogger(__name__)


class Memory(object):
    @staticmethod
    def rss_to_mb(rss_in_bytes):
        """
        Convert a number of bytes to a number of megabytes.
        """
        bytes_per_kb = 1024
        kb_per_megabyte = 1024
        return float(rss_in_bytes) / bytes_per_kb / kb_per_megabyte

    @staticmethod
    def rss():
        """
        Returns the memory usage (RSS) of this process, in bytes
        """
        return psutil.Process().memory_info().rss

    @staticmethod
    def rss_in_mb():
        return Memory.rss_to_mb(Memory.rss())

    def metric_type(self):
        return 'Memory'

    def metric_name(self):
        return 'Physical'

    def human_name(self):
        return 'Process Memory'

    def run(self):
        res = self.__class__.rss_in_mb()
        logger.debug('{human_name}: #{res}'.format(
            human_name=self.human_name(),
            res=res))
        return res



from __future__ import absolute_import

import threading


class ThreadLocalSingleton(object):
    @classmethod
    def instance(cls, *args, **kwargs):
        if not hasattr(cls, '_thread_lookup'):
            cls.__new_instance(args, kwargs)
        elif not hasattr(cls._thread_lookup, 'instance'):
            cls.__new_instance(args, kwargs)
        elif cls._thread_lookup.instance is None:
            cls.__new_instance(args, kwargs)
        return cls._thread_lookup.instance

    @classmethod
    def __new_instance(cls, *args, **kwargs):
        cls._thread_lookup = threading.local()
        cls._thread_lookup.instance = cls(args, kwargs)

    def release(self):
            if getattr(self.__class__._thread_lookup,
                       'instance',
                       None) is self:
                self.__class__._thread_lookup.instance = None



import scout_apm.django
import scout_apm.django.apps


def test_default_django_config():
    assert(scout_apm.django.default_app_config == 'scout_apm.django.apps.ScoutApmDjangoConfig')


def test_apps():
    assert(scout_apm.django.apps.ScoutApmDjangoConfig.name == 'scout_apm')



from __future__ import absolute_import
import logging

from .config.config import ScoutConfig
from .socket import CoreAgentSocket
from .thread_local import ThreadLocalSingleton

# Logging
logger = logging.getLogger(__name__)


class AgentContext(ThreadLocalSingleton):
    def __init__(self, *args, **kwargs):
        self.config = kwargs.get('config', ScoutConfig())

    @classmethod
    def socket(cls):
        return CoreAgentSocket.instance(scout_config=ScoutConfig())



#!/usr/bin/env python
from __future__ import absolute_import
import argparse
import logging


def download(**kwargs):
    from scout_apm.core import CoreAgentManager
    core_agent_manager = CoreAgentManager()
    core_agent_manager.download()


def launch(**kwargs):
    from scout_apm.core import CoreAgentManager
    core_agent_manager = CoreAgentManager()
    core_agent_manager.launch()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="count")

    subparsers = parser.add_subparsers(dest='subparser')

    download_parser = subparsers.add_parser('download')

    launch_parser = subparsers.add_parser('launch')

    args = parser.parse_args()

    if args.verbose is not None:
        if args.verbose >= 2:
            logging.basicConfig(level=getattr(logging, 'DEBUG', None))
        elif args.verbose == 1:
            logging.basicConfig(level=getattr(logging, 'INFO', None))

    kwargs = vars(args)
    globals()[kwargs.pop('subparser')](**kwargs)



from __future__ import absolute_import

# Python Built-Ins
import hashlib
import json
import logging
import os
import subprocess
import tarfile
import time

# 3rd Party
import requests

# APM Modules
from scout_apm.core.context import AgentContext

# Logging
logger = logging.getLogger(__name__)


class CoreAgentManager:
    def __init__(self):
        self.core_agent_bin_path = None
        self.core_agent_bin_version = None
        self.core_agent_dir = '{}/{}'.format(AgentContext.instance().config.value('core_agent_dir'),
                                             AgentContext.instance().config.core_agent_full_name())
        self.downloader = CoreAgentDownloader(self.core_agent_dir,
                                              AgentContext.instance().config.core_agent_full_name())

    def launch(self):
        if AgentContext.instance().config.value('core_agent_launch') is not True:
            logger.debug("Not attempting to launch Core Agent "
                         "due to 'core_agent_launch' setting.")
            return

        if self.verify() is not True:
            if AgentContext.instance().config.value('core_agent_download') is True:
                self.download()
            else:
                logger.debug("Not attempting to download Core Agent due "
                             "to 'core_agent_download' setting.")

        if self.verify() is not True:
            logger.debug("Failed to verify Core Agent. "
                         "Not launching Core Agent.")
            return False

        return self.run()

    def download(self):
        self.downloader.download()

    def run(self):
        try:
            process = subprocess.Popen(
                    self.agent_binary() +
                    self.daemonize_flag() +
                    self.log_level() +
                    self.log_file() +
                    self.config_file() +
                    self.socket_path()
                    )
            process.wait()
        except Exception as e:
            # TODO detect failute of launch properly
            logger.error('Error running Core Agent: %s', repr(e))
            return False
        return True

    def agent_binary(self):
        return [self.core_agent_bin_path, 'start']

    def daemonize_flag(self):
        return ['--daemonize', 'true']

    def socket_path(self):
        socket_path = AgentContext.instance().config.value('socket_path')
        return ['--socket', socket_path]

    def log_level(self):
        level = AgentContext.instance().config.value('log_level')
        return ['--log-level', level]

    def log_file(self):
        path = AgentContext.instance().config.value('log_file')
        if path is not None:
            return ['--log-file', path]
        else:
            return []

    def config_file(self):
        path = AgentContext.instance().config.value('config_file')
        if path is not None:
            return ['--config-file', path]
        else:
            return []

    def verify(self):
        manifest = CoreAgentManifest(self.core_agent_dir + '/manifest.json')
        if manifest.is_valid() is not True:
            logger.debug('Core Agent verification failed: '
                         'CoreAgentManifest is not valid.')
            self.core_agent_bin_path = None
            self.core_agent_bin_version = None
            return False

        bin_path = self.core_agent_dir + '/' + manifest.bin_name
        if SHA256.digest(bin_path) == manifest.sha256:
            self.core_agent_bin_path = bin_path
            self.core_agent_bin_version = manifest.bin_version
            return True
        else:
            logger.debug('Core Agent verification failed: SHA mismatch.')
            self.core_agent_bin_path = None
            self.core_agent_bin_version = None
            return False


class CoreAgentDownloader():
    def __init__(self, download_destination, core_agent_full_name):
        self.stale_download_secs = 120
        self.destination = download_destination
        self.core_agent_full_name = core_agent_full_name
        self.package_location = self.destination + '/{}.tgz'.format(self.core_agent_full_name)
        self.download_lock_path = self.destination + '/download.lock'
        self.download_lock_fd = None

    def download(self):
        self.create_core_agent_dir()
        self.obtain_download_lock()
        if self.download_lock_fd is not None:
            try:
                self.download_package()
                self.untar()
            except OSError as e:
                logger.error('Exception raised while '
                             'downloading Core Agent: %s', repr(e))
            finally:
                self.release_download_lock()

    def create_core_agent_dir(self):
        try:
            os.makedirs(self.destination, 0o700)
        except OSError:
            pass

    def obtain_download_lock(self):
        self.clean_stale_download_lock()
        try:
            self.download_lock_fd = os.open(self.download_lock_path, os.O_RDWR | os.O_CREAT | os.O_EXCL | os.O_NONBLOCK)
        except OSError as e:
            logger.debug("Could not obtain download lock on %s: %s",
                         self.download_lock_path,
                         repr(e))
            self.download_lock_fd = None

    def clean_stale_download_lock(self):
        try:
            delta = time.time() - os.stat(self.download_lock_path).st_ctime
            if delta > self.stale_download_secs:
                logger.debug("Clearing stale download lock file.")
                os.unlink(self.download_lock_path)
        except OSError:
            pass

    def release_download_lock(self):
        if self.download_lock_fd is not None:
            os.unlink(self.download_lock_path)
            os.close(self.download_lock_fd)

    def download_package(self):
        logger.debug('Downloading: {full_url} to {filepath}'.format(
            full_url=self.full_url(),
            filepath=self.package_location))
        req = requests.get(self.full_url(), stream=True)
        with open(self.package_location, 'wb') as f:
            for chunk in req.iter_content(1024 * 1000):
                f.write(chunk)

    def untar(self):
        t = tarfile.open(self.package_location, 'r')
        t.extractall(self.destination)

    def full_url(self):
        return '{root_url}/{core_agent_full_name}.tgz'.format(
                root_url=self.root_url(),
                core_agent_full_name=self.core_agent_full_name)

    def root_url(self):
        return AgentContext.instance().config.value('download_url')


class CoreAgentManifest:
    def __init__(self, path):
        self.manifest_path = path
        self.bin_name = None
        self.bin_version = None
        self.sha256 = None
        self.valid = False
        try:
            self.parse()
        except (ValueError, TypeError, OSError) as e:
            logger.debug('Error parsing Core Agent Manifest: %s', repr(e))

    def parse(self):
        logger.debug('Parsing Core Agent'
                     ' manifest path: %s', self.manifest_path)
        with open(self.manifest_path) as manifest_file:
            self.raw = manifest_file.read()
            self.json = json.loads(self.raw)
            self.version = self.json['version']
            self.bin_version = self.json['core_agent_version']
            self.bin_name = self.json['core_agent_binary']
            self.sha256 = self.json['core_agent_binary_sha256']
            self.valid = True
            logger.debug("Core Agent manifest json: %s", self.json)

    def is_valid(self):
        return self.valid


class SHA256:
    @staticmethod
    def digest(filename, block_size=65536):
        try:
            sha256 = hashlib.sha256()
            with open(filename, 'rb') as f:
                for block in iter(lambda: f.read(block_size), b''):
                    sha256.update(block)
            return sha256.hexdigest()
        except OSError as e:
            logger.debug('Error on digest: %s', repr(e))
            return None



from __future__ import absolute_import

import logging
from datetime import datetime
from uuid import uuid4

from scout_apm.core.samplers import Samplers
from scout_apm.core.request_manager import RequestManager
from scout_apm.core.thread_local import ThreadLocalSingleton

# Logging
logger = logging.getLogger(__name__)


class TrackedRequest(ThreadLocalSingleton):
    """
    This is a container which keeps track of all module instances for a single
    request. For convenience they are made available as attributes based on
    their keyname
    """
    def __init__(self, *args, **kwargs):
        self.req_id = 'req-' + str(uuid4())
        self.start_time = kwargs.get('start_time', datetime.utcnow())
        self.end_time = kwargs.get('end_time', None)
        self.active_spans = kwargs.get('active_spans', [])
        self.complete_spans = kwargs.get('complete_spans', [])
        self.tags = kwargs.get('tags', {})
        self.real_request = kwargs.get('real_request', False)
        logger.debug('Starting request: %s', self.req_id)

    def mark_real_request(self):
        self.real_request = True

    def is_real_request(self):
        return self.real_request

    def tag(self, key, value):
        if hasattr(self.tags, key):
            logger.debug('Overwriting previously set tag for request %s: %s' % self.req_id, key)
        self.tags[key] = value

    def start_span(self, operation=None):
        maybe_parent = self.current_span()

        if maybe_parent is not None:
            parent_id = maybe_parent.span_id
        else:
            parent_id = None

        new_span = Span(
            request_id=self.req_id,
            operation=operation,
            parent=parent_id)
        self.active_spans.append(new_span)
        return new_span

    def stop_span(self):
        stopping_span = self.active_spans.pop()
        stopping_span.stop()
        self.complete_spans.append(stopping_span)
        if len(self.active_spans) == 0:
            self.finish()

    def current_span(self):
        if len(self.active_spans) > 0:
            return self.active_spans[-1]
        else:
            return None

    # Request is done, release any info we have about it.
    def finish(self):
        logger.debug('Stopping request: %s', self.req_id)
        if self.end_time is None:
            self.end_time = datetime.utcnow()
        RequestManager.instance().add_request(self)
        if self.is_real_request():
            Samplers.ensure_running()
        self.release()


class Span:
    def __init__(self, *args, **kwargs):
        self.span_id = kwargs.get('span_id', 'span-' + str(uuid4()))
        self.start_time = kwargs.get('start_time', datetime.utcnow())
        self.end_time = kwargs.get('end_time', None)
        self.request_id = kwargs.get('request_id', None)
        self.operation = kwargs.get('operation', None)
        self.parent = kwargs.get('parent', None)
        self.tags = kwargs.get('tags', {})

    def dump(self):
        if self.end_time is None:
            logger.debug(self.operation)
        return 'request=%s operation=%s id=%s parent=%s start_time=%s end_time=%s' % (
                self.request_id,
                self.operation,
                self.span_id,
                self.parent,
                self.start_time.isoformat(),
                self.end_time.isoformat()
            )

    def stop(self):
        self.end_time = datetime.utcnow()

    def tag(self, key, value):
        if hasattr(self.tags, key):
            logger.debug('Overwriting previously set tag for span %s: %s' % self.span_id, key)
        self.tags[key] = value

    # In seconds
    def duration(self):
        if self.end_time is not None:
            (self.end_time - self.start_time).total_seconds()
        else:
            # Current, running duration
            (datetime.utcnow() - self.start_time).total_seconds()



from django.conf import settings
from scout_apm.core.config.config import ScoutConfig


import logging


logger = logging.getLogger(__name__)


class ConfigAdapter:
    @classmethod
    def install(cls):
        configs = {}
        for name in filter(lambda x: x.startswith('SCOUT_'), dir(settings)):
            value = getattr(settings, name)
            clean_name = name.replace('SCOUT_', '').lower()
            configs[clean_name] = value
        ScoutConfig.set(**configs)



# pylint: skip-file

from __future__ import absolute_import

# Originally taken from https://pypi.python.org/pypi/ProxyTypes
# inlined due to python3 issues with setup.py, we should gut this out and simplify it


class AbstractProxy(object):
    """Delegates all operations (except ``.__subject__``) to another object"""
    __slots__ = ()

    def __call__(self, *args, **kw):
        return self.__subject__(*args, **kw)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        subject = oga(self, '__subject__')
        if attr == '__subject__':
            return subject
        return getattr(subject, attr)

    def __setattr__(self, attr, val, osa=object.__setattr__):
        if attr == '__subject__':
            osa(self, attr, val)
        else:
            setattr(self.__subject__, attr, val)

    def __delattr__(self, attr, oda=object.__delattr__):
        if attr == '__subject__':
            oda(self, attr)
        else:
            delattr(self.__subject__, attr)

    def __nonzero__(self):
        return bool(self.__subject__)

    def __getitem__(self, arg):
        return self.__subject__[arg]

    def __setitem__(self, arg, val):
        self.__subject__[arg] = val

    def __delitem__(self, arg):
        del self.__subject__[arg]

    def __getslice__(self, i, j):
        return self.__subject__[i:j]

    def __setslice__(self, i, j, val):
        self.__subject__[i:j] = val

    def __delslice__(self, i, j):
        del self.__subject__[i:j]

    def __contains__(self, ob):
        return ob in self.__subject__

    for name in 'repr str hash len abs complex int long float iter oct hex'.split():
        exec('def __%s__(self): return %s(self.__subject__)' % (name, name))

    for name in 'cmp', 'coerce', 'divmod':
        exec('def __%s__(self,ob): return %s(self.__subject__,ob)' % (name, name))

    for name, op in [
        ('lt', '<'), ('gt', '>'), ('le', '<='), ('ge', '>='),
        ('eq', '=='), ('ne', '!=')
    ]:
        exec('def __%s__(self,ob): return self.__subject__ %s ob' % (name, op))

    for name, op in [('neg', '-'), ('pos', '+'), ('invert', '~')]:
        exec('def __%s__(self): return %s self.__subject__' % (name, op))

    for name, op in [
        ('or', '|'),  ('and', '&'), ('xor', '^'), ('lshift', '<<'),
        ('rshift', '>>'), ('add', '+'), ('sub', '-'), ('mul', '*'),
        ('div', '/'), ('mod', '%'), ('truediv', '/'), ('floordiv', '//')
    ]:
        exec((
            'def __%(name)s__(self,ob):\n'
            '    return self.__subject__ %(op)s ob\n'
            '\n'
            'def __r%(name)s__(self,ob):\n'
            '    return ob %(op)s self.__subject__\n'
            '\n'
            'def __i%(name)s__(self,ob):\n'
            '    self.__subject__ %(op)s=ob\n'
            '    return self\n'
        ) % locals())

    del name, op

    # Oddball signatures

    def __rdivmod__(self, ob):
        return divmod(ob, self.__subject__)

    def __pow__(self, *args):
        return pow(self.__subject__, *args)

    def __ipow__(self, ob):
        self.__subject__ **= ob
        return self

    def __rpow__(self, ob):
        return pow(ob, self.__subject__)


class ObjectProxy(AbstractProxy):
    """Proxy for a specific object"""

    __slots__ = '__subject__'

    def __init__(self, subject):
        self.__subject__ = subject


class CallbackProxy(AbstractProxy):
    """Proxy for a dynamically-chosen object"""

    __slots__ = '__callback__'

    def __init__(self, func):
        set_callback(self, func)


set_callback = CallbackProxy.__callback__.__set__
get_callback = CallbackProxy.__callback__.__get__
CallbackProxy.__subject__ = property(lambda self, gc=get_callback: gc(self)())


class LazyProxy(CallbackProxy):
    """Proxy for a lazily-obtained object, that is cached on first use"""
    __slots__ = '__cache__'


get_cache = LazyProxy.__cache__.__get__
set_cache = LazyProxy.__cache__.__set__


def __subject__(self, get_cache=get_cache, set_cache=set_cache):
    try:
        return get_cache(self)
    except AttributeError:
        set_cache(self, get_callback(self)())
        return get_cache(self)


LazyProxy.__subject__ = property(__subject__, set_cache)
del __subject__


class AbstractWrapper(AbstractProxy):
    """Mixin to allow extra behaviors and attributes on proxy instance"""
    __slots__ = ()

    def __getattribute__(self, attr, oga=object.__getattribute__):
        if attr.startswith('__'):
            subject = oga(self, '__subject__')
            if attr == '__subject__':
                return subject
            return getattr(subject, attr)
        return oga(self, attr)

    def __getattr__(self, attr, oga=object.__getattribute__):
        return getattr(oga(self, '__subject__'), attr)

    def __setattr__(self, attr, val, osa=object.__setattr__):
        if (
            attr == '__subject__' or
            hasattr(type(self), attr) and not attr.startswith('__')
        ):
            osa(self, attr, val)
        else:
            setattr(self.__subject__, attr, val)

    def __delattr__(self, attr, oda=object.__delattr__):
        if (
            attr == '__subject__' or
            hasattr(type(self), attr) and not attr.startswith('__')
        ):
            oda(self, attr)
        else:
            delattr(self.__subject__, attr)


class ObjectWrapper(ObjectProxy, AbstractWrapper):
    __slots__ = ()


class CallbackWrapper(CallbackProxy, AbstractWrapper):
    __slots__ = ()


class LazyWrapper(LazyProxy, AbstractWrapper):
    __slots__ = ()


#######################
#  MONKEYS DOWN HERE  #
#######################


# pylint: disable=E1001,E1002
class CallableProxy(ObjectWrapper):
    __slots__ = ('_eop_wrapper_')

    def __init__(self, wrapped, wrapper):
        super(CallableProxy, self).__init__(wrapped)
        self._eop_wrapper_ = wrapper

    def __call__(self, *args, **kwargs):
        return self._eop_wrapper_(self.__subject__, *args, **kwargs)


class BoundMethodProxy(ObjectWrapper):
    __slots__ = ('_eop_wrapper_', '_eop_instance_')

    def __init__(self, wrapped, instance, wrapper):
        super(BoundMethodProxy, self).__init__(wrapped)
        self._eop_instance_ = instance
        self._eop_wrapper_ = wrapper

    def __call__(self, *args, **kwargs):
        return self._eop_wrapper_(self.__subject__, self._eop_instance_, *args, **kwargs)


class UnboundMethodProxy(CallableProxy):
    __slots__ = ('_eop_wrapper_')

    def __get__(self, instance, owner):
        return BoundMethodProxy(self.__subject__.__get__(instance, owner), instance or owner, self._eop_wrapper_)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        """
        We need to return our own version of __get__ or we may end up
        never being called if the member we are wrapping is wrapped
        again by someone else
        """
        if attr == '__get__':
            return oga(self, attr)
        return super(UnboundMethodProxy, self).__getattribute__(attr)


def monkeypatch_method(cls, method_name=None):
    def decorator(func):
        method_to_patch = method_name or func.__name__
        original = cls.__dict__[method_to_patch]
        replacement = UnboundMethodProxy(original, func)
        type.__setattr__(cls, method_to_patch, replacement)  # Avoid overrides
        return func
    return decorator



import scout_apm
import os
import re


def test_defaults():
    conf = scout_apm.core.config.config.ScoutConfig()
    m = re.match(r'/tmp/scout_apm_core/scout_apm_core-latest-(linux|darwin|unknown)-(x86_64|i686|unknown)/core-agent\.sock',
                 conf.value('socket_path'))
    assert(m is not None)


def test_env():
    conf = scout_apm.core.config.config.ScoutConfig()
    os.environ['SCOUT_SOCKET_PATH'] = '/set/in/env'
    assert('/set/in/env' == conf.value('socket_path'))
    del os.environ['SCOUT_SOCKET_PATH']


def test_python():
    scout_apm.core.config.config.ScoutConfig.set(socket_path='/set/via/function')
    conf = scout_apm.core.config.config.ScoutConfig()
    assert('/set/via/function' == conf.value('socket_path'))


def test_none():
    conf = scout_apm.core.config.config.ScoutConfig()
    assert(conf.value('unknown value') is None)


def test_env_outranks_python():
    os.environ['SCOUT_SOCKET_PATH'] = '/set/in/env'
    scout_apm.core.config.config.ScoutConfig.set(socket_path='/set/via/function')

    conf = scout_apm.core.config.config.ScoutConfig()
    assert('/set/in/env' == conf.value('socket_path'))

    del os.environ['SCOUT_SOCKET_PATH']




from __future__ import absolute_import

# Python Modules
from datetime import datetime
import json
import logging

logger = logging.getLogger(__name__)


class Register:
    def __init__(self, *args, **kwargs):
        self.app = kwargs.get('app', None)
        self.key = kwargs.get('key', None)

    def message(self):
        logging.info('Registering with app=%s key=%s' % (self.app, self.key))
        return {'Register': {
            'app': self.app,
            'key': self.key,
            'api_version': '1.0',
        }}


class StartSpan:
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.request_id = kwargs.get('request_id', None)
        self.span_id = kwargs.get('span_id', None)
        self.parent = kwargs.get('parent', None)
        self.operation = kwargs.get('operation', None)

    def message(self):
        return {'StartSpan': {
            'timestamp': self.timestamp.isoformat() + 'Z',
            'request_id': self.request_id,
            'span_id': self.span_id,
            'parent_id': self.parent,
            'operation': self.operation,
        }}


class StopSpan:
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.request_id = kwargs.get('request_id', None)
        self.span_id = kwargs.get('span_id', None)

    def message(self):
        return {'StopSpan': {
            'timestamp': self.timestamp.isoformat() + 'Z',
            'request_id': self.request_id,
            'span_id': self.span_id,
        }}


class StartRequest:
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.request_id = kwargs.get('request_id', None)

    def message(self):
        return {'StartRequest': {
            'timestamp': self.timestamp.isoformat() + 'Z',
            'request_id': self.request_id,
        }}


class FinishRequest:
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.request_id = kwargs.get('request_id', None)

    def message(self):
        return {'FinishRequest': {
            'timestamp': self.timestamp.isoformat() + 'Z',
            'request_id': self.request_id,
        }}


class TagSpan:
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.request_id = kwargs.get('request_id', None)
        self.span_id = kwargs.get('span_id', None)
        self.tag = kwargs.get('tag', None)
        self.value = kwargs.get('value', None)

    def message(self):
        return {'TagSpan': {
            'timestamp': self.timestamp.isoformat() + 'Z',
            'request_id': self.request_id,
            'span_id': self.span_id,
            'tag': self.tag,
            'value': self.value,
        }}


class TagRequest:
    def __init__(self, *args, **kwargs):
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.request_id = kwargs.get('request_id', None)
        self.tag = kwargs.get('tag', None)
        self.value = kwargs.get('value', None)

    def message(self):
        return {'TagRequest': {
            'timestamp': self.timestamp.isoformat() + 'Z',
            'request_id': self.request_id,
            'tag': self.tag,
            'value': self.value,
        }}


class ApplicationEvent:
    def __init__(self, *args, **kwargs):
        self.event_type = kwargs.get('event_type', '')
        self.event_value = kwargs.get('event_value', '')
        self.timestamp = kwargs.get('timestamp', datetime.utcnow())
        self.source = kwargs.get('source', '')

    def message(self):
        return {'ApplicationEvent': {
                    'event_type':  self.event_type,
                    'event_value': self.event_value,
                    'timestamp': self.timestamp.isoformat() + 'Z',
                    'source': self.source,
        }}


class CoreAgentVersion:
    def message(self):
        return {'CoreAgentVersion': {}}


class CoreAgentVersionResponse:
    def __init__(self, message):
        self.loaded = json.loads(message)
        self.version = self.loaded['CoreAgentVersion']['version']


class BatchCommand:
    def __init__(self, commands):
        self.commands = commands

    def message(self):
        messages = list(map(lambda cmd: cmd.message(), self.commands))
        return {'BatchCommand': {'commands': messages}}

    @classmethod
    def from_tracked_request(cls, request):
        commands = []
        # Request Start
        commands.append(StartRequest(timestamp=request.start_time,
                                     request_id=request.req_id))
        # Request Tags
        for key in request.tags:
            commands.append(TagRequest(timestamp=request.start_time,
                                       request_id=request.req_id,
                                       tag=key,
                                       value=request.tags[key]))
        # Spans
        for span in request.complete_spans:
            # Span Start
            commands.append(StartSpan(timestamp=span.start_time,
                                      request_id=span.request_id,
                                      span_id=span.span_id,
                                      parent=span.parent,
                                      operation=span.operation))
            # Span Tags
            for key in span.tags:
                commands.append(TagSpan(timestamp=span.start_time,
                                        request_id=request.req_id,
                                        span_id=span.span_id,
                                        tag=key,
                                        value=span.tags[key]))
            # Span End
            commands.append(StopSpan(timestamp=span.end_time,
                                     request_id=span.request_id,
                                     span_id=span.span_id))
        # Request Finish
        commands.append(FinishRequest(timestamp=request.end_time,
                                      request_id=request.req_id))
        return BatchCommand(commands)



from __future__ import absolute_import

# Python Modules
from datetime import datetime
import logging
from os import getpid
import pip
import sys


# Scout APM
from scout_apm.core.context import AgentContext
from scout_apm.core.commands import ApplicationEvent
# from scout_apm.environment import Environment

logger = logging.getLogger(__name__)


class AppMetadata():
    @classmethod
    def report(cls):
        event = ApplicationEvent()
        event.event_value = cls.data()
        event.event_type = 'scout.metadata'
        event.timestamp = datetime.utcnow()
        event.source = 'Pid: ' + str(getpid())
        AgentContext.socket().send(event)

    @classmethod
    def data(cls):
        data = {}
        version_tuple = sys.version_info
        try:
            data = {'language':          'python',
                    'version':           '{}.{}.{}'.format(version_tuple[0],
                                                           version_tuple[1],
                                                           version_tuple[2]),
                    'server_time':        datetime.utcnow().isoformat() + 'Z',
                    'framework':          '',
                    'framework_version':  '',
                    'environment':        '',
                    'app_server':         '',
                    'hostname':           '',  # Environment.hostname,
                    'database_engine':    '',  # Detected
                    'database_adapter':   '',  # Raw
                    'application_name':   '',  # Environment.application_name,
                    'libraries':          cls.package_list(),
                    'paas':               '',
                    'git_sha':            ''}  # Environment.git_revision_sha()}
        except Exception as e:
            logger.debug('Exception in AppMetadata: %s', repr(e))

        return data

    @classmethod
    def package_list(cls):
        packages = []
        try:
            for pkg_dist in [p for p in pip._vendor.pkg_resources.working_set]:
                try:
                    p_split = str(pkg_dist).split()
                    packages.append([p_split[0], p_split[1]])
                except Exception as e:
                    logger.debug('Exception while reading packages: %s', repr(e))
                    continue
        except Exception as e:
            logger.debug('Exception while getting pkg_resources: %s', repr(e))
        return packages



from __future__ import absolute_import

import logging
import os
import platform

logger = logging.getLogger(__name__)


class ScoutConfig():
    """
    Configuration object for the ScoutApm agent.

    Contains a list of configuration "layers". When a configuration key is
    looked up, each layer is asked in turn if it knows the value. The first one
    to answer affirmatively returns the value.
    """
    def __init__(self):
        self.layers = [
            ScoutConfigEnv(),
            ScoutConfigPython(),
            ScoutConfigDefaults(),
            ScoutConfigNull()]

    def value(self, key):
        value = self.locate_layer_for_key(key).value(key)
        if key in CONVERSIONS:
            converted_value = CONVERSIONS[key].convert(value)
        else:
            converted_value = value

        return converted_value

    def locate_layer_for_key(self, key):
        for layer in self.layers:
            if layer.has_config(key):
                return layer

    def log(self):
        logger.debug('Configuration Loaded:')
        for key in self.known_keys():
            layer = self.locate_layer_for_key(key)
            logger.debug('{:9}: {} = {}'.format(
                layer.name(),
                key,
                layer.value(key)))

    def known_keys(self):
        return [
            'core_agent_dir',
            'core_agent_download',
            'core_agent_launch',
            'core_agent_version',
            'download_url',
            'git_sha',
            'log_level',
            'name',
            'key',
            'socket_path'
        ]

    def core_agent_full_name(self):
        return 'scout_apm_core-{version}-{platform}-{arch}'.format(
                version=self.value('core_agent_version'),
                platform=self.platform(),
                arch=self.arch())

    @classmethod
    def platform(cls):
        system_name = platform.system()
        if system_name == 'Linux':
            return 'linux'
        elif system_name == 'Darwin':
            return 'darwin'
        else:
            return 'unknown'

    @classmethod
    def arch(cls):
        arch = platform.machine()
        if arch == 'i686':
            return 'i686'
        elif arch == 'x86_64':
            return 'x86_64'
        else:
            return 'unknown'

    @classmethod
    def set(cls, **kwargs):
        """
        Sets a configuration value for the Scout agent. Values set here will
        not override values set in ENV.
        """
        global SCOUT_PYTHON_VALUES
        for key, value in kwargs.items():
            SCOUT_PYTHON_VALUES[key] = value


# Module-level data, the ScoutConfig.set(key="value") adds to this
SCOUT_PYTHON_VALUES = {}


class ScoutConfigPython():
    """
    A configuration overlay that lets other parts of python set values.
    """
    def name(self):
        return 'Python'

    def has_config(self, key):
        return key in SCOUT_PYTHON_VALUES

    def value(self, key):
        return SCOUT_PYTHON_VALUES[key]



class ScoutConfigEnv():
    """
    Reads configuration from environment by prefixing the key
    requested with "SCOUT_"

    Example: the `log_level` config looks for SCOUT_LOG_LEVEL
    environment variable
    """

    def name(self):
        return 'ENV'

    def has_config(self, key):
        env_key = self.modify_key(key)
        return env_key in os.environ

    def value(self, key):
        env_key = self.modify_key(key)
        return os.environ[env_key]

    def modify_key(self, key):
        env_key = ('SCOUT_' + key).upper()
        return env_key


class ScoutConfigDefaults():
    """
    Provides default values for important configurations
    """

    def name(self):
        return 'Defaults'

    def __init__(self):
        self.core_agent_dir = '/tmp/scout_apm_core'
        self.core_agent_version = 'latest'
        self.defaults = {
                'core_agent_dir': self.core_agent_dir,
                'core_agent_download': True,
                'core_agent_launch': True,
                'core_agent_version': self.core_agent_version,
                'download_url': 'https://s3-us-west-1.amazonaws.com/scout-public-downloads/apm_core_agent/release',
                'git_sha': '',
                'key': '',
                'log_level': 'info',
                'name': '',
                'monitor': False,
                'socket_path': '{}/scout_apm_core-{}-{}-{}/core-agent.sock'.format(self.core_agent_dir,
                                                                                   self.core_agent_version,
                                                                                   ScoutConfig.platform(),
                                                                                   ScoutConfig.arch())
        }

    def has_config(self, key):
        return key in self.defaults

    def value(self, key):
        return self.defaults[key]


# Always returns None to any key
class ScoutConfigNull():
    """
    Always answers that a key is present, but the value is None

    Used as the last step of the layered configuration.
    """

    def name(self):
        return 'Null'

    def has_config(self, key):
        return True

    def value(self, key):
        return None


class BooleanConversion():
    @classmethod
    def convert(cls, value):
        if isinstance(value, bool):
            return value
        if isinstance(value, str):
            return value.lower() in ('yes', 'true', 't', '1')
        # Unknown type - default to false?
        return False


CONVERSIONS = {
    'core_agent_download': BooleanConversion,
    'core_agent_launch': BooleanConversion,
    'monitor': BooleanConversion,
}



import logging
from datetime import datetime, timedelta
from textwrap import dedent

import psutil

# Logging
logger = logging.getLogger(__name__)


class Cpu(object):
    def __init__(self):
        self.last_run = datetime.now()
        self.last_cpu_times = psutil.Process().cpu_times()
        self.num_processors = psutil.cpu_count()

    def metric_type(self):
        return "CPU"

    def metric_name(self):
        return "Utilization"

    def human_name(self):
        return "Process CPU"

    def run(self):
        now = datetime.now()
        process = psutil.Process() # get a handle on the current process
        cpu_times = process.cpu_times()

        wall_clock_elapsed = now - self.last_run
        if wall_clock_elapsed < timedelta(0):
            self.save_times(now, cpu_times)
            logger.debug(dedent(
                         """
                         {human_name}: Negative time elapsed.
                         now: {now},
                         last_run: {last_run},
                         total time: {wall_clock_elapsed}.
                         """
                         ).format(human_name=self.human_name(),
                                  now=now,
                                  last_run=self.last_run))
            return None

        utime_elapsed = cpu_times.user - self.last_cpu_times.user
        stime_elapsed = cpu_times.system - self.last_cpu_times.system
        process_elapsed = utime_elapsed + stime_elapsed

        # This can happen right after a fork.  This class starts up in
        # pre-fork, records {u,s}time, then forks. This resets {u,s}time to 0
        if process_elapsed < 0:
            self.save_times(now, cpu_times)
            logger.debug(dedent(
                         """
                         {human_name}: Negative process time elapsed.
                         utime: {utime_elapsed},
                         stime: {stime_elapsed},
                         total time: {process_elapsed}.
                         This is normal to see when starting a
                         forking web server.
                         """
                        ).format(human_name=self.human_name(),
                                 utime_elapsed=utime_elapsed,
                                 stime_elapsed=stime_elapsed,
                                 process_elapsed=process_elapsed))
            return None

        # Normalized to # of processors
        normalized_wall_clock_elapsed = (wall_clock_elapsed * self.num_processors).total_seconds()

        # If somehow we run for 0 seconds between calls, don't try to divide by 0
        res = None
        if normalized_wall_clock_elapsed == 0:
            res = 0
        else:
            res = (process_elapsed / normalized_wall_clock_elapsed)*100

        if res < 0:
            self.save_times(now, cpu_times)
            logger.debug(dedent(
                         """
                         {human_name}: Negative CPU.
                         {process_elapsed} / {normalized_wall_clock_elapsed} * 100 ==> {res}
                         """
                         ).format(human_name=self.human_name(),
                                  process_elapsed=process_elapsed,
                                  normalized_wall_clock_elapsed=normalized_wall_clock_elapsed,
                                  res=res))
            return None

        self.save_times(now, cpu_times)

        logger.debug(dedent(
                     """
                     {human_name}: {res} [{num_processors} CPU(s)]
                     """
                     ).format(human_name=self.human_name(),
                              res=res,
                              num_processors=self.num_processors))

        return res

    def save_times(self, now, cpu_times):
        self.last_run = now
        self.cpu_times = cpu_times



from __future__ import absolute_import
import logging

from scout_apm.core.stacktracer import trace_function, trace_method

# XXX:Changed in Django 1.9
# https://github.com/jazzband/django-debug-toolbar/issues/739
#  from django.template.base import Library, Template
from django.template import Library, Template, defaulttags
from django.template.loader_tags import BlockNode

logger = logging.getLogger(__name__)

register = Library()


class DecoratingParserProxy(object):
    """
    Mocks out the django template parser, passing templatetags through but
    first wrapping them to include performance data
    """
    def __init__(self, parser):
        self.parser = parser

    def add_library(self, library):
        wrapped_library = Library()
        wrapped_library.filters = library.filters
        for name, tag_compiler in library.tags.items():
            wrapped_library.tags[name] = self.wrap_compile_function(name, tag_compiler)
        self.parser.add_library(wrapped_library)

    def wrap_compile_function(self, name, tag_compiler):
        def compile(*args, **kwargs):
            node = tag_compiler(*args, **kwargs)
            node.render = trace_function(node.render, ('Template/Tag', {"name": name}))
            return node
        return compile


@register.tag
def load(parser, token):
    decorating_parser = DecoratingParserProxy(parser)
    return defaulttags.load(decorating_parser, token)


class TemplateInstrument:
    # The linter thinks the methods we monkeypatch are not used
    # pylint: disable=W0612
    @staticmethod
    def install():
        # Our eventual aim is to patch the render() method on the Node objects
        # corresponding to custom template tags. However we have to jump through
        # a number of hoops in order to get access to the object.
        #
        # 1. Add ourselves to the set of built in template tags
        #    This allows us to replace the 'load' template tag which controls
        #    the loading of custom template tags
        # 2. Delegate to default load with replaced parser
        #    We provide our own parser class so we can catch and intercept
        #    calls to add_library.
        # 3. add_library receives a library of template tags
        #    It iterates through each template tag, wrapping its compile function
        # 4. compile is called as part of compiling the template
        #    Our wrapper is called instead of the original templatetag compile
        #    function. We delegate to the original function, but then modify
        #    the resulting object by wrapping its render() function. This
        #    render() function is what ends up being timed and appearing in the
        #    tree.

        # XXX: Stopped working in Django 1.9
        #  add_to_builtins('apm.instruments.view')

        @trace_method(Template)
        def __init__(self, *args, **kwargs):
            name = args[2] if len(args) >= 3 else '<Unknown Template>'
            return ('Template/Compile', {'name': name})

        @trace_method(Template)
        def render(self, *args, **kwargs):
            name = self.name if self.name is not None else '<Unknown Template>'
            return ('Template/Render', {'name': name})

        @trace_method(BlockNode)
        def render(self, *args, **kwargs):
            return ('Block/Render', {'name': self.name})

        logger.debug('Monkey patched Templates')

        # XXX: Figure this out, causes exception that the "resolve_context" key isn't in dict
        # Also will need to figure out the name hash
        #  @trace_method(TemplateResponse)
        #  def resolve_context(self, *args, **kwargs):
        #  return ('Template/Context', {})



from __future__ import absolute_import
import logging

from scout_apm.core.monkey import CallableProxy, monkeypatch_method
from scout_apm.core.tracked_request import TrackedRequest

# Logging
logger = logging.getLogger(__name__)


# The linter thinks the methods we monkeypatch are not used
# pylint: disable=W0612
def trace_method(cls, method_name=None):
    def decorator(info_func):
        method_to_patch = method_name or info_func.__name__

        @monkeypatch_method(cls, method_to_patch)
        def tracing_method(original, self, *args, **kwargs):
            entry_type, detail = info_func(self, *args, **kwargs)

            operation = entry_type
            if detail['name'] is not None:
                operation = operation + '/' + detail['name']

            tr = TrackedRequest.instance()
            span = tr.start_span(operation=operation)

            for key in detail:
                span.tag(key, detail[key])

            try:
                return original(*args, **kwargs)
            finally:
                TrackedRequest.instance().stop_span()
                logger.debug(span.dump())
        return tracing_method
    return decorator


def trace_function(func, info):
    try:
        def tracing_function(original, *args, **kwargs):
            if callable(info):
                entry_type, detail = info(*args, **kwargs)
            else:
                entry_type, detail = info

            operation = entry_type
            if detail['name'] is not None:
                operation = operation + '/' + detail['name']

            tr = TrackedRequest.instance()
            span = tr.start_span(operation=operation)

            for key in detail:
                span.tag(key, detail[key])

            try:
                return original(*args, **kwargs)
            finally:
                TrackedRequest.instance().stop_span()
                logger.debug(span.dump())

        return CallableProxy(func, tracing_function)
    except Exception:
        # If we can't wrap for any reason, just return the original
        return func



from __future__ import absolute_import

import json
import logging
import socket
import struct
import time
import threading

try:
    # Python 3.x
    import queue
except ImportError:
    # Python 2.x
    import Queue as queue

from scout_apm.core.config.config import ScoutConfig
from scout_apm.core.commands import Register

# Logging
logger = logging.getLogger(__name__)


class CoreAgentSocket(threading.Thread):
    _instance = None
    _run_lock = threading.Semaphore()

    @classmethod
    def instance(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = cls(args, kwargs)
        elif cls._instance.running() is False:
            del(cls._instance)
            cls._instance = cls(args, kwargs)
        return cls._instance

    def __init__(self, *args, **kwargs):
        # Call threading.Thread.__init__()
        super(CoreAgentSocket, self).__init__()
        self.config = kwargs.get('scout_config', ScoutConfig())
        # Socket related
        self.socket_path = self.config.value('socket_path')
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        # Threading control related
        self._started_event = threading.Event()
        self._stop_event = threading.Event()
        self._stopped_event = threading.Event()
        # Command queues
        self.command_queue = queue.Queue(maxsize=500)
        # Start the thread
        self.daemon = True
        self.start()

    def __del__(self):
        self.stop()

    def running(self):
        return self._started_event.is_set()

    def stop(self):
        if self._started_event.is_set():
            self._stop_event.set()
            self._stopped_event.wait(2)
            if self._stopped_event.is_set():
                return True
            else:
                logger.debug('CoreAgentSocket Failed to stop thread within timeout!')
                return False
        else:
            return True

    def run(self):
        if self.__class__._run_lock.acquire(False) is False:
            logger.debug('CoreAgentSocket thread failed to acquire run lock.')
            return None

        try:
            self._started_event.set()
            self._connect()
            self._register()
            while True:
                try:
                    body = self.command_queue.get(block=True, timeout=1)
                except queue.Empty:
                    continue

                if body is not None:
                    result = self._send(body)
                    if result is True:
                        self.command_queue.task_done()
                    else:
                        # Something was wrong with the socket.
                        self._disconnect()
                        self._connect()
                        self._register()

                # Check for stop event after a read from the queue. This is to
                # allow you to open a socket, immediately send to it, and then
                # stop it. We do this in the Metadata send at application start
                # time
                if self._stop_event.is_set():
                    logger.debug("CoreAgentSocket thread stopping.")
                    break
        finally:
            self.__class__._run_lock.release()
            self._stop_event.clear()
            self._started_event.clear()
            self._stopped_event.set()
            logger.debug("CoreAgentSocket thread stopped.")

    def send(self, command):
        try:
            self.command_queue.put(command, False)
        except queue.Full as e:
            # TODO mark the command as not queued?
            logger.debug('CoreAgentSocket error on send: %s' % repr(e))

    def _send(self, command, async=True):
        msg = command.message()

        try:
            data = json.dumps(msg)
        except (ValueError, TypeError) as e:
            log.debug('Exception when serializing command message: %s' % repr(e))
            return False

        try:
            self.socket.sendall(self._message_length(data))
            self.socket.sendall(data.encode())
        except (OSError, ConnectionRefusedError) as e:
            logger.debug("CoreAgentSocket exception on _send: %s" % repr(e))
            return None

        if async is True:
            return True
        else:
            # TODO read respnse back in to command
            self._read_response()
            return True

    def _message_length(self, body):
        length = len(body)
        return length.to_bytes(4, 'big')

    def _read_response(self):
        try:
            raw_size = self.socket.recv(4)
            size = struct.unpack('<I', raw_size)[0]
            message = self.socket.recv(size)
            return message
        except (OSError, ConnectionRefusedError) as e:
            logger.debug('CoreAgentSocket error on read response: %s' % repr(e))
            return None

    def _register(self):
        self._send(Register(app=self.config.value('name'),
                            key=self.config.value('key')))

    def _connect(self, connect_attempts=5, retry_wait_secs=1):
        for attempt in range(1, connect_attempts):
            logger.debug('CoreAgentSocket attempt %d, connecting to %s', attempt, self.socket_path)
            try:
                self.socket.connect(self.socket_path)
                self.socket.settimeout(0.5)
                logger.debug('CoreAgentSocket is connected')
                return True
            except (FileNotFoundError, ConnectionRefusedError) as e:
                logger.debug('CoreAgentSocket connection error: %s', repr(e))
                if attempt >= connect_attempts:
                    return False
                time.sleep(retry_wait_secs)
                continue

    def _disconnect(self):
        logger.debug('CoreAgentSocket disconnecting from %s', self.socket_path)
        try:
            self.socket.close()
        except (OSError, ConnectionRefusedError) as e:
            logger.debug('CoreAgentSocket exception on disconnect: %s' % repr(e))
        finally:
            self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)



from __future__ import absolute_import
import logging

from scout_apm.core.monkey import monkeypatch_method
from scout_apm.core.tracked_request import TrackedRequest

try:
    from django.db.backends.base.base import BaseDatabaseWrapper
except ImportError:
    # Backwards compatibility for Django <1.8
    from django.db.backends import BaseDatabaseWrapper

try:
    from django.db.backends.utils import CursorWrapper
except ImportError:
    # Backwards compatibility for Django <1.9
    from django.db.backends.util import CursorWrapper

logger = logging.getLogger(__name__)


class _DetailedTracingCursorWrapper(CursorWrapper):
    def execute(self, sql, params=()):
        tr = TrackedRequest.instance()
        span = tr.start_span(operation='SQL/Query')
        span.tag('db.statement', sql)

        try:
            return self.cursor.execute(sql, params)
        finally:
            tr.stop_span()

    def executemany(self, sql, param_list):
        span = TrackedRequest.instance().start_span(operation='SQL/Many')
        span.tag('db.statement', sql)

        try:
            return self.cursor.executemany(sql, param_list)
        finally:
            TrackedRequest.instance().stop_span()


# pylint: disable=too-few-public-methods
class SQLInstrument:

    # The linter thinks the methods we monkeypatch are not used
    # pylint: disable=W0612
    # pylint: disable=no-method-argument
    @staticmethod
    def install():
        """
        Installs ScoutApm SQL Instrumentation by monkeypatching the `cursor`
        method of BaseDatabaseWrapper, to return a wrapper that instruments any
        calls going through it.
        """
        @monkeypatch_method(BaseDatabaseWrapper)
        def cursor(original, self, *args, **kwargs):
            result = original(*args, **kwargs)
            return _DetailedTracingCursorWrapper(result, self)

        logger.debug('Monkey patched SQL')



from __future__ import absolute_import

import logging

from scout_apm.core.context import AgentContext
from scout_apm.core.thread_local import ThreadLocalSingleton

from .commands import BatchCommand

# Logging
logger = logging.getLogger(__name__)


class RequestManager(ThreadLocalSingleton):
    def __init__(self, *args, **kwargs):
        self.request_buffer = RequestBuffer()

    def add_request(self, request):
        if request.is_real_request():
            self.request_buffer.add_request(request)


class RequestBuffer(ThreadLocalSingleton):
    def __init__(self):
        self._requests = []

    # TODO: ensure there is a limit to the tracked requests in this buffer
    def add_request(self, request):
        self._requests.append(request)
        self.flush()

    def flush(self):
        logger.debug('Flushing RequestBuffer')
        for request in self._requests:
            logger.debug('Flushing Request Id: %s' % request.req_id)
            self.flush_request(request)
        self._requests.clear()

    def flush_request(self, request):
        batch_command = BatchCommand.from_tracked_request(request)
        AgentContext.socket().send(batch_command)



from datetime import datetime
import logging
import threading
from os import getpid
from time import sleep

from scout_apm.core.context import AgentContext
from scout_apm.core.commands import ApplicationEvent

from scout_apm.core.samplers.cpu import Cpu
from scout_apm.core.samplers.memory import Memory

# Logging
logger = logging.getLogger(__name__)


class Samplers():
    _thread_lock = threading.Semaphore()

    @classmethod
    def ensure_running(cls):
        if cls._thread_lock.acquire(False) is True:
            th = threading.Thread(target=Samplers.run_samplers)
            th.daemon = True
            th.start()
            cls._thread_lock.release()

    @classmethod
    def run_samplers(cls):
        logger.debug('Starting Samplers. Acquiring samplers lock.')
        try:
            if cls._thread_lock.acquire(True) is True:
                logger.debug('Acquired samplers lock.')
                instances = [Cpu(), Memory()]

                while True:
                    for instance in instances:
                        event = ApplicationEvent()
                        event.event_value = instance.run()
                        event.event_type = instance.metric_type() + '/' + instance.metric_name()
                        event.timestamp = datetime.utcnow()
                        event.source = 'Pid: ' + str(getpid())

                        if event.event_value is not None:
                            AgentContext.socket().send(event)
                    sleep(60)
        finally:
            logger.debug('Shutting down samplers thread.')
            cls._thread_lock.release()



from __future__ import absolute_import


# pip needs to be imported before anything else (in particular the requests
# library), since it vendors stuff oddly.
#
# This is an unsupported use of pip, and we need to figure out a smoother way
# to detect loaded libraries. Check metadata.py for our usage.
import pip # noqa

# Python Modules
import logging
from os import getpid

from scout_apm.core.context import AgentContext
from scout_apm.core.core_agent_manager import CoreAgentManager
from scout_apm.core.metadata import AppMetadata

logger = logging.getLogger(__name__)


def install():
    if not AgentContext.instance().config.value("monitor"):
        logger.debug('APM Not Launching on PID: %s - Configuration \'monitor\' is not true', getpid())
        return False

    logger.debug('APM Launching on PID: %s', getpid())
    CoreAgentManager().launch()
    AppMetadata.report()
    AgentContext.socket().stop()



from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup

setup(name='scout_apm',
      version='1.0.0',
      description='Scout Application Performance Monitoring Agent',
      long_description='Scout Application Performance Monitoring Agent',
      url='https://github.com/scoutapp/scout_apm_python',
      author='Scout',
      author_email='support@scoutapp.com',
      license='Proprietary',
      zip_safe=False,
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
      entry_points={
          'console_scripts': [
              'core-agent-manager = scout_apm.cli.core_agent_manager:__main__'
          ]
      },
      install_requires=['psutil', 'PyYAML', 'requests'],
      keywords='apm performance monitoring development',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: System :: Monitoring',
        'License :: Other/Proprietary License',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
      ])
